@@ -22,7 +22,7 @@ struct options opt;
 
 const struct option_info options_info[] = {
   { C_R, OPT_SETUIDGID,   'u',  NULL,       required_argument, "set uid, gid and supplementary groups", "[:]USER[:GROUP]*", },
-  { C_R, OPT_ENVUIDGID,   'U',  NULL,       required_argument, "set UID and GID vars", "[:]USER[:GROUP]", },
+  { C_R, OPT_ENVUIDGID,   'U',  NULL,       required_argument, "set UID, GID and GIDLIST vars", "[:]USER[:GROUP]*", },
   { C_R, OPT_ARGV0,       'b',  NULL,       required_argument, "launch program with ARGV0 as the argv[0]", "ARGV0" },
   { C_R, OPT_ENVDIR,      'e',  NULL,       required_argument, "populate environment from directory", "DIR" },
   { C_R, OPT_CHROOT,      '/',  NULL,       required_argument, "change root directory", "DIR" },
@@ -74,10 +74,12 @@ const struct option_info options_info[] = {
   { C_X, OPT_RO_HOME,     '\0', "ro-home",      no_argument,    "make home directories read only", NULL },
   { C_X, OPT_RO_SYS,      '\0', "ro-sys",       no_argument,    "create read only system", NULL },
   { C_X, OPT_RO_ETC,      '\0', "ro-etc",       no_argument,    "make config files read only", NULL },
+#ifndef NO_CAP
   { C_X, OPT_CAPBS_KEEP,  '\0', "cap-bs-keep",  required_argument, "restrict capabilities bounding set", "CAP[,...]" },
   { C_X, OPT_CAPBS_DROP,  '\0', "cap-bs-drop",  required_argument, "drop from capabilities bounding set", "CAP[,...]" },
   { C_X, OPT_CAPS_KEEP,   '\0', "caps-keep",    required_argument, "keep (only) these capabilities", "CAP[,...]" },
   { C_X, OPT_CAPS_DROP,   '\0', "caps-drop",    required_argument, "drop these capabilities", "CAP[,...]" },
+#endif
   { C_X, OPT_FORK_JOIN,   '\0', "fork-join",    no_argument,   "fork and wait for process", NULL },
   { C_X, OPT_NEW_ROOT,    '\0', "new-root",     no_argument,   "create a new root fs", NULL },
   { C_X, OPT_NO_NEW_PRIVS,'\0', "no-new-privs", no_argument,   "no new privileges", NULL },
@@ -93,6 +95,13 @@ const struct option_info options_info[] = {
   { C_X, OPT_LOGIN,       '\0', "login",     no_argument,      "simulate login environment", NULL },
   { C_X, OPT_OOM,         '\0', "oom",       required_argument,"set oom adjust value", "ADJ" },
   { C_X, OPT_HARDLIMIT,   '\0', "hardlimit", no_argument,      "set hard limits with soft limits", NULL },
+  { C_A, OPT_UIDNUM,      'u',  NULL,        required_argument,"set numerical uid", "UID" },
+  { C_A, OPT_GIDNUM,      'g',  NULL,        required_argument,"set numerical gid", "GID" },
+  { C_A, OPT_GIDSNUM,     'G',  NULL,        required_argument,"set numerical supplementary groups", "GID*" },
+  { C_A, OPT_UGFROMENV,   'U',  NULL,              no_argument,"set numerical uid, gid and supplementary groups from env", NULL },
+  { C_X, OPT_UGFROMENV,   '\0', "ugids-from-env",  no_argument,"set numerical uid, gid and supplementary groups from env", NULL },
+  { C_A, OPT_UGCLEARENV,  'z',  NULL,              no_argument,"clear numerical uid, gid and supplementary groups in env", NULL },
+  { C_X, OPT_UGCLEARENV,  '\0', "ugids-clear-env", no_argument,"clear numerical uid, gid and supplementary groups in env", NULL },
 };
 #define max_options ((ssize_t) ((sizeof options_info / sizeof *options_info)))
 
@@ -278,6 +287,7 @@ bool parse_limits(struct limit *limit, const char *arg) {
   return limit->soft_specified || limit->hard_specified;
 }
 
+#ifndef NO_CAP
 bool parse_caps(cap_bits_t *caps, char *names) {
   cap_bits_t set = 0;
   cap_value_t val;
@@ -305,6 +315,7 @@ bool parse_caps(cap_bits_t *caps, char *names) {
   *caps = set;
   return good;
 }
+#endif
 
 bool parse_cpu_range(const char *str, int range[3], int *max) {
   int field = 0;
@@ -707,8 +718,10 @@ static void handle_option(enum compat_level *compat,
   case OPT_LOG_DIR:
   case OPT_LOGIN:
   case OPT_HARDLIMIT:
+  case OPT_UGCLEARENV:
     /* Boolean options needing no further option processing */
     break;
+#ifndef NO_CAP
   case OPT_CAPBS_KEEP:
     if (!parse_caps(&opt.cap_bounds, optarg))
       opt.error = true;
@@ -729,6 +742,7 @@ static void handle_option(enum compat_level *compat,
       opt.error = true;
     opt.caps_op = CAP_OP_DROP;
     break;
+#endif
   case OPT_CPU_SCHED:
     opt.sched_policy = sched_policy_from_name(optarg);
     break;
@@ -766,6 +780,33 @@ static void handle_option(enum compat_level *compat,
     if (opt.verbosity > 1)
       usrgrp_print(stderr, "envuidgid", &opt.env_users_groups);
     break;
+  case OPT_UIDNUM:
+    if (usrgrp_from_env(&opt.users_groups, optarg, NULL, NULL))
+      opt.error = true;
+    else
+      enable(OPT_SETUIDGID);
+    break;
+  case OPT_GIDNUM:
+    if (usrgrp_from_env(&opt.users_groups, NULL, optarg, NULL))
+      opt.error = true;
+    else
+      enable(OPT_SETUIDGID);
+    break;
+  case OPT_GIDSNUM:
+    if (usrgrp_from_env(&opt.users_groups, NULL, NULL, optarg))
+      opt.error = true;
+    else
+      enable(OPT_SETUIDGID);
+    break;
+  case OPT_UGFROMENV:
+    if (usrgrp_from_env(&opt.users_groups,
+                        getenv("UID"), getenv("GID"), getenv("GIDLIST")))
+      opt.error = true;
+    else
+      enable(OPT_SETUIDGID);
+    if (opt.verbosity > 1)
+      usrgrp_print(stderr, "ugfromenv", &opt.users_groups);
+    break;
   case OPT_LIMIT_MEM:
     if (!parse_limits(&opt.rlimit_memlock, optarg)) {
       opt.error = true;
@@ -896,6 +937,7 @@ int options_parse(int argc, char *argv[]) {
          optdef++);
     assert(optdef);
 
+    enable(optdef->option);
     handle_option(&compat, optdef,
                   optdef->has_arg == no_argument ? NULL : argv[optind++]);
   }
