@@ -52,17 +52,25 @@ static const char *default_app = "xchpst";
 #define PROG "xchpst"
 
 const struct app apps[] = {
-  { COMPAT_CHPST,     "chpst",     .long_opts = false },
-  { COMPAT_XCHPST,    "xchpst",    .long_opts = true },
-  { COMPAT_SOFTLIMIT, "softlimit", .long_opts = false },
-  { COMPAT_ENVDIR,    "envdir",    false, 1, { OPT_ENVDIR } },
-  { COMPAT_PGRPHACK,  "pgrphack",  false, 1, { OPT_PGRPHACK } },
-  { COMPAT_SETUIDGID, "setuidgid", false, 1, { OPT_SETUIDGID } },
-  { COMPAT_ENVUIDGID, "envuidgid", false, 1, { OPT_ENVUIDGID } },
-  { COMPAT_SETLOCK,   "setlock",   false, 1, { OPT_LOCK_WAIT } },
+  { COMPAT_CHPST,       "chpst",     .long_opts = false },
+  { COMPAT_XCHPST,      "xchpst",    .long_opts = true },
+  { COMPAT_SOFTLIMIT,   "softlimit", .long_opts = false },
+  { COMPAT_ENVDIR,      "envdir",    false, 1, { OPT_ENVDIR } },
+  { COMPAT_PGRPHACK,    "pgrphack",  false, 1, { OPT_PGRPHACK } },
+  { COMPAT_SETUIDGID,   "setuidgid", false, 1, { OPT_SETUIDGID } },
+  { COMPAT_ENVUIDGID,   "envuidgid", false, 1, { OPT_ENVUIDGID } },
+  { COMPAT_SETLOCK,     "setlock",   false, 1, { OPT_LOCK_WAIT } },
+  { COMPAT_APPLYUIDGID, "applyuidgid", .long_opts = false },
+  { COMPAT_SETUIDGID_FROMENV,
+                        "setuidgid-fromenv", false, 1, { OPT_UGFROMENV } },
 };
 #define max_apps ((ssize_t) (sizeof apps / (sizeof *apps)))
 
+static const char *suite_prefixes[] = {
+  "s6-",
+  NULL
+};
+
 struct runtime runtime = { 0 };
 
 static const char *std_run_dir = "/run/" PROG;
@@ -73,7 +81,8 @@ char *run_dir;
 static struct {
   char uid[16];
   char gid[16];
-} extra_env;
+  char *gidlist;
+} extra_env = { 0 };
 
 static void version(FILE *out) {
   fprintf(out,
@@ -210,24 +219,28 @@ void set_resource_limits(void) {
   set_rlimit(RLIMIT_LOCKS, &opt.rlimit_locks);
 }
 
-bool drop_user(uid_t uid, gid_t gid) {
+bool drop_user(struct users_groups *ug) {
   int rc;
   int i;
+  uid_t uid = ug->user.uid;
+  gid_t gid = ug->group.gid;
 
   /* You do this backwards: supplemental groups first */
-  gid_t *groups = malloc(sizeof(gid_t) * opt.users_groups.num_supplemental);
+  gid_t *groups = malloc(sizeof(gid_t) * ug->num_supplemental);
 
+#ifndef NO_CAP
   if (opt.caps_op != CAP_OP_NONE) {
     /* Postpone the loss of capabilities until after the user switch
      * and drop them when ready. */
     rc = prctl(PR_SET_KEEPCAPS, 1);
   }
+#endif
 
-  if (!groups)
+  if (!groups && ug->num_supplemental)
     return false;
 
-  for (i = 0; i < opt.users_groups.num_supplemental; i++)
-    groups[i] = opt.users_groups.supplemental[i].gid;
+  for (i = 0; i < ug->num_supplemental; i++)
+    groups[i] = ug->supplemental[i].gid;
   rc = (setgroups(i, groups) == -1 ? errno : 0);
   free(groups);
   if (rc) {
@@ -236,15 +249,15 @@ bool drop_user(uid_t uid, gid_t gid) {
   }
 
   /* Then main group */
-  rc = setresgid(gid, gid, gid);
-  if (rc) {
+  if (ug->group.resolved &&
+      (rc = setresgid(gid, gid, gid))) {
     perror("setresgid");
     return false;
   }
 
   /* Then the actual user */
-  rc = setresuid(uid, uid, uid);
-  if (rc) {
+  if (ug->user.resolved &&
+      (rc = setresuid(uid, uid, uid))) {
     perror("setresgid");
     return false;
   }
@@ -254,6 +267,11 @@ bool drop_user(uid_t uid, gid_t gid) {
 static const struct app *find_app(const char *name) {
   const struct app *app;
   const char *ext = strchrnul(name, '.');
+  const char **suite;
+
+  for (suite = suite_prefixes; *suite && strncmp(name, *suite, strlen(*suite)); suite++);
+  if (*suite)
+    name += strlen(*suite);
 
   for (app = apps;
        app - apps < max_apps && strncmp(name, app->name, ext - name);
@@ -267,8 +285,7 @@ int main(int argc, char *argv[]) {
   sigset_t oldmask;
   char **sub_argv;
   char *executable;
-  char *new_root = NULL;
-  char *old_root = NULL;
+  struct pivot_state pivot = pivot_init();
   int sub_argc;
   pid_t child;
   int optind;
@@ -385,23 +402,51 @@ int main(int argc, char *argv[]) {
     goto finish;
 
   if (set(OPT_ENVUIDGID)) {
+    struct users_groups *ug = &opt.env_users_groups;
     struct sys_entry *entry;
-    if ((entry = &opt.env_users_groups.user)->resolved) {
+    const ssize_t idsz = 12;
+    char *s;
+    int i;
+    if ((entry = &ug->user)->resolved) {
       rc = snprintf(extra_env.uid, sizeof extra_env.uid, "UID=%d", entry->uid);
       if (rc == -1 || rc == sizeof extra_env.uid) {
         perror("creating UID env");
         goto finish;
       }
       putenv(extra_env.uid);
+    } else {
+      unsetenv("UID");
+      fprintf(stderr, "envuidgid: no user resolved\n");
+      goto finish;
     }
-    if ((entry = &opt.env_users_groups.group)->resolved) {
+    if ((entry = &ug->group)->resolved) {
       rc = snprintf(extra_env.gid, sizeof extra_env.gid, "GID=%d", entry->gid);
       if (rc == -1 || rc == sizeof extra_env.uid) {
-        perror("creating UID env");
+        perror("creating GID env");
         goto finish;
       }
       putenv(extra_env.gid);
+    } else {
+      unsetenv("GID");
     }
+    if ((s = extra_env.gidlist = malloc(1 + idsz * ug->num_supplemental)) == NULL)
+      goto finish;
+    for (i = 0; i < ug->num_supplemental; i++) {
+      rc = snprintf(s, idsz, "%d,", ug->supplemental[i].gid);
+      if (rc < 0 || rc >= idsz) {
+        perror("creating GIDLIST env");
+        goto finish;
+      }
+      s += rc;
+    }
+    s[i ? -1 : 0] = '\0';
+    setenv("GIDLIST", extra_env.gidlist, 1);
+  }
+
+  if (set(OPT_UGCLEARENV)) {
+    unsetenv("UID");
+    unsetenv("GID");
+    unsetenv("GIDLIST");
   }
 
   if (set(OPT_NICE)) {
@@ -437,6 +482,7 @@ int main(int argc, char *argv[]) {
       sched_setscheduler(0, opt.sched_policy, &((struct sched_param) {})) == -1)
     perror("could not change scheduler policy");
 
+#ifndef NO_CAP
   if ((opt.cap_bounds_op != CAP_OP_NONE ||
        opt.caps_op != CAP_OP_NONE) &&
       runtime.absent.caps) {
@@ -447,12 +493,22 @@ int main(int argc, char *argv[]) {
   if (opt.cap_bounds_op != CAP_OP_NONE)
     if (!set_capabilities_bounding_set())
       goto finish;
+#endif
 
-  if (set(OPT_SETUIDGID) && opt.users_groups.user.resolved) {
-    gid = opt.users_groups.group.gid;
-    uid = opt.users_groups.user.uid;
+  if (set(OPT_SETUIDGID)) {
+    if (opt.users_groups.user.resolved) {
+      uid = opt.users_groups.user.uid;
+    } else {
+      fprintf(stderr, "setuidgid: no user resolved\n");
+      goto finish;
+    }
   } else {
     uid = getuid();
+  }
+
+  if (set(OPT_SETUIDGID) && opt.users_groups.group.resolved) {
+    gid = opt.users_groups.group.gid;
+  } else {
     gid = getgid();
   }
 
@@ -469,7 +525,7 @@ int main(int argc, char *argv[]) {
     else if (set(OPT_ENVUIDGID))
       ug = &opt.env_users_groups;
     else
-      usrgrp_resolve_uid(ug = &current, getuid());
+      usrgrp_resolve_uid(ug = &current, getuid(), true);
 
     if (ug->user.resolved) {
       if (ug->username && *ug->username) {
@@ -519,8 +575,7 @@ int main(int argc, char *argv[]) {
   /* Iff using a user namespace, drop the user first */
   if (set(OPT_SETUIDGID) &&
       opt.new_ns & CLONE_NEWUSER &&
-      opt.users_groups.user.resolved &&
-      !drop_user(uid, gid))
+      !drop_user(&opt.users_groups))
       goto finish;
 
   if (opt.new_ns) {
@@ -573,7 +628,7 @@ int main(int argc, char *argv[]) {
   }
 
   if (set(OPT_NEW_ROOT)) {
-    if (!create_new_root(basename(executable), &new_root, &old_root))
+    if (!create_new_root(basename(executable), &pivot))
       goto finish;
   }
 
@@ -611,7 +666,7 @@ int main(int argc, char *argv[]) {
    *************************************/
 
   if (set(OPT_NEW_ROOT)) {
-    if (!pivot_to_new_root(new_root, old_root))
+    if (!pivot_to_new_root(&pivot))
       goto finish;
     else
       in_new_root = true;
@@ -677,13 +732,14 @@ int main(int argc, char *argv[]) {
 
   if (set(OPT_SETUIDGID) &&
       (opt.new_ns & CLONE_NEWUSER) == 0 &&
-      opt.users_groups.user.resolved &&
-      !drop_user(uid, gid))
+      !drop_user(&opt.users_groups))
       goto finish;
 
+#ifndef NO_CAP
   if (opt.caps_op != CAP_OP_NONE)
     if (!drop_capabilities())
       goto finish;
+#endif
 
   for (unsigned int close_fds = opt.close_fds; close_fds; close_fds &= ~(1 << fd))
     close(fd = /*stdc_trailing_zeros*/ __builtin_ctz(close_fds));
@@ -691,6 +747,8 @@ int main(int argc, char *argv[]) {
   if (set(OPT_NO_NEW_PRIVS) && prctl(PR_SET_NO_NEW_PRIVS, 1L, 0L, 0L, 0L) == -1)
     perror("could not honour --no-new-privs");
 
+  pivot_tidy(&pivot);
+
   /* Launch the target */
   rc = execvp(executable, sub_argv);
 
@@ -709,24 +767,20 @@ finish:
      3) not be necessary when --fork-join is not used.
    */
 
-  if (new_root && !in_new_root) {
-    if (umount2(new_root, MNT_DETACH) == -1)
-      fprintf(stderr, "umount2(%s): %s\n", new_root, strerror(errno));
-    if (rmdir(new_root) == -1)
-      fprintf(stderr, "rmdir(%s): %s\n", new_root, strerror(errno));
+  if (pivot.path && !in_new_root) {
     free_rootfs_data();
   }
 
+  pivot_tidy(&pivot);
   if (run_dir_fd != -1)
     close(run_dir_fd);
 
   if (lock_fd != -1)
     close(lock_fd);
 
-  free(new_root);
-  free(old_root);
   free(sub_argv);
   free(run_dir);
+  free(extra_env.gidlist);
 
 finish0:
   options_free();
